from .auto_ai import AutoAI
from .easy_ai import EasyAI


# Making certain symbols available when the package is imported
__all__ = ['AutoAI', 'EasyAI']
#print(f"Initializing ai package, available classes: {__all__}")
